"""Define Pydantic validators."""
from __future__ import annotations

from datetime import datetime


def validate_timestamp(value: str | None) -> datetime | None:
    """Validate a timestamp.

    Args:
        value: An ISO 8601-formatted timestamp string.

    Returns:
        A parsed datetime.datetime object (UTC).
    """
    if not value:
        return None
    return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z")



"""Define system models."""
# pylint: disable=consider-alternative-union-syntax,too-few-public-methods
from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import BaseModel, validator

from aionotion.helpers.validators import validate_timestamp


class System(BaseModel):
    """Define a system."""

    uuid: str
    name: str
    mode: str
    partners: list[str]
    latitude: float
    longitude: float
    timezone_id: str
    created_at: datetime
    updated_at: datetime
    night_time_start: datetime
    night_time_end: datetime
    id: int
    locality: str
    postal_code: str
    administrative_area: str
    fire_number: str
    police_number: str
    emergency_number: str
    address: Optional[str]
    notion_pro_permit: Optional[str]

    validate_created_at = validator("created_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )
    validate_updated_at = validator("updated_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )
    validate_night_time_start = validator(
        "night_time_start", allow_reuse=True, pre=True
    )(validate_timestamp)
    validate_night_time_end = validator("night_time_end", allow_reuse=True, pre=True)(
        validate_timestamp
    )


class SystemAllResponse(BaseModel):
    """Define an API response containing all systems."""

    systems: list[System]


class SystemGetResponse(BaseModel):
    """Define an API response containing a single system."""

    system: System

    class Config:
        """Define model configuration."""

        fields = {
            "system": "systems",
        }



"""Define a base client for interacting with Notion."""
from __future__ import annotations

from typing import Any

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError
from pydantic import ValidationError

from aionotion.bridge import Bridge
from aionotion.const import LOGGER
from aionotion.device import Device
from aionotion.errors import InvalidCredentialsError, RequestError
from aionotion.helpers.typing import BaseModelT
from aionotion.sensor import Sensor
from aionotion.system import System

API_BASE: str = "https://api.getnotion.com/api"

DEFAULT_TIMEOUT: int = 10


class Client:  # pylint: disable=too-few-public-methods
    """Define the API object."""

    def __init__(self, *, session: ClientSession | None = None) -> None:
        """Initialize.

        Args:
            session: An optional aiohttp ClientSession.
        """
        self._session = session
        self._token: str | None = None

        self.bridge = Bridge(self._request, self._request_and_validate)
        self.device = Device(self._request, self._request_and_validate)
        self.sensor = Sensor(self._request, self._request_and_validate)
        self.system = System(self._request, self._request_and_validate)

    async def _request(
        self, method: str, endpoint: str, **kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        """Make an API request.

        Args:
            method: An HTTP method.
            endpoint: A relative API endpoint.
            **kwargs: Additional kwargs to send with the request.

        Returns:
            An API response payload.

        Raises:
            InvalidCredentialsError: Raised upon invalid credentials.
            RequestError: Raised upon an underlying HTTP error.
        """
        url: str = f"{API_BASE}/{endpoint}"

        kwargs.setdefault("headers", {})
        if self._token:
            kwargs["headers"]["Authorization"] = f"Token token={self._token}"

        if use_running_session := self._session and not self._session.closed:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        data: dict[str, Any] = {}

        async with session.request(method, url, **kwargs) as resp:
            data = await resp.json()

            try:
                resp.raise_for_status()
            except ClientError as err:
                if "401" in str(err):
                    raise InvalidCredentialsError("Invalid credentials") from err
                raise RequestError(data["errors"][0]["title"]) from err

        if not use_running_session:
            await session.close()

        LOGGER.debug("Received data from /%s: %s", endpoint, data)

        return data

    async def _request_and_validate(
        self,
        method: str,
        endpoint: str,
        model: BaseModelT,
        **kwargs: dict[str, Any],
    ) -> BaseModelT:
        """Make an API request and validate the response against a Pydantic model.

        Args:
            method: An HTTP method.
            endpoint: A relative API endpoint.
            model: A Pydantic model to validate the response against.
            **kwargs: Additional kwargs to send with the request.

        Returns:
            A parsed, validated Pydantic model representing the response.
        """
        raw_data = await self._request(method, endpoint, **kwargs)

        try:
            return model.parse_obj(raw_data)
        except ValidationError as err:
            raise RequestError(
                f"Error while parsing response from {endpoint}: {err}"
            ) from err

    async def async_authenticate(self, email: str, password: str) -> None:
        """Authenticate the user and retrieve an authentication token.

        Args:
            email: The email address of a Notion account.
            password: The account password.
        """
        auth_response = await self._request(
            "post",
            "users/sign_in",
            json={"sessions": {"email": email, "password": password}},
        )

        self._token = auth_response["session"]["authentication_token"]


async def async_get_client(
    email: str, password: str, *, session: ClientSession | None = None
) -> Client:
    """Return an authenticated API object.

    Args:
        email: The email address of a Notion account.
        password: The account password.
        session: An optional aiohttp ClientSession.

    Returns:
        An authenticated Client object.
    """
    client: Client = Client(session=session)
    await client.async_authenticate(email, password)
    return client



"""Define sensor models."""
# pylint: disable=consider-alternative-union-syntax,too-few-public-methods
from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Literal, Optional

from pydantic import BaseModel, Extra, validator

from aionotion.const import LOGGER
from aionotion.helpers.validators import validate_timestamp


class Bridge(BaseModel):
    """Define a bridge representation."""

    id: int
    hardware_id: str


class Firmware(BaseModel):
    """Define firmware information."""

    status: str


class SurfaceType(BaseModel):
    """Define a surface type."""

    id: str
    name: str
    slug: str


class User(BaseModel):
    """Define a user representation."""

    id: int
    email: str


class Sensor(BaseModel):
    """Define a sensor."""

    id: int
    uuid: str
    user: User
    bridge: Bridge
    last_bridge_hardware_id: str
    name: str
    location_id: int
    system_id: int
    hardware_id: str
    hardware_revision: int
    firmware_version: str
    device_key: str
    encryption_key: bool
    installed_at: datetime
    calibrated_at: datetime
    last_reported_at: datetime
    missing_at: Optional[datetime]
    updated_at: datetime
    created_at: datetime
    signal_strength: int
    firmware: Firmware
    surface_type: Optional[SurfaceType]

    validate_installed_at = validator("installed_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )
    validate_calibrated_at = validator("calibrated_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )
    validate_last_reported_at = validator(
        "last_reported_at", allow_reuse=True, pre=True
    )(validate_timestamp)
    validate_missing_at = validator("missing_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )
    validate_created_at = validator("created_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )
    validate_updated_at = validator("updated_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )


class SensorAllResponse(BaseModel):
    """Define an API response containing all sensors."""

    sensors: list[Sensor]


class SensorGetResponse(BaseModel):
    """Define an API response containing a single sensor."""

    sensor: Sensor

    class Config:
        """Define model configuration."""

        fields = {
            "sensor": "sensors",
        }


class ListenerStatus(BaseModel, extra=Extra.allow):
    """Define a listener status."""

    trigger_value: str
    data_received_at: datetime

    validate_data_received_at = validator(
        "data_received_at", allow_reuse=True, pre=True
    )(validate_timestamp)


class ListenerLocalizedStatus(BaseModel):
    """Define a localized listener status."""

    state: str
    description: str


class InsightOrigin(BaseModel):
    """Define an insight origin."""

    type: Optional[str]
    id: Optional[str]


class PrimaryListenerInsight(BaseModel):
    """Define a primary listener insight."""

    origin: Optional[InsightOrigin]
    value: Optional[str]
    data_received_at: Optional[datetime]

    validate_data_received_at = validator(
        "data_received_at", allow_reuse=True, pre=True
    )(validate_timestamp)


class ListenerInsights(BaseModel):
    """Define listener insights:"""

    primary: PrimaryListenerInsight


class ListenerKind(Enum):
    """Define the kinds of listener."""

    BATTERY = 0
    MOLD = 2
    TEMPERATURE = 3
    LEAK_STATUS = 4
    SAFE = 5
    DOOR = 6
    SMOKE = 7
    CONNECTED = 10
    HINGED_WINDOW = 12
    GARAGE_DOOR = 13
    SLIDING_DOOR_OR_WINDOW = 32
    UNKNOWN = 99


class Listener(BaseModel):
    """Define a listener."""

    id: str
    listener_kind: ListenerKind
    created_at: datetime
    device_type: str
    model_version: str
    sensor_id: str
    status: ListenerStatus
    status_localized: ListenerLocalizedStatus
    insights: ListenerInsights
    configuration: dict[str, Any]
    pro_monitoring_status: Literal["eligible", "ineligible"]

    @validator("listener_kind", pre=True)
    @classmethod
    def validate_listener_kind(cls, value: str) -> ListenerKind:
        """Validate the API key type.

        Args:
            value: An API key to validate.

        Returns:
            A parsed ApiKeyType.

        Raises:
            ValueError: An invalid API key type was received.
        """
        try:
            return ListenerKind(value)
        except ValueError:
            LOGGER.error("Received an unknown listener kind: %s", value)
            return ListenerKind.UNKNOWN

    validate_created_at = validator("created_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )

    class Config:
        """Define model configuration."""

        fields = {
            "device_type": "type",
            "listener_kind": "definition_id",
        }


class ListenerAllResponse(BaseModel):
    """Define an API response containing all listeners."""

    listeners: list[Listener]



"""Define device models."""
# pylint: disable=too-few-public-methods
from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel, validator

from aionotion.helpers.validators import validate_timestamp


class Device(BaseModel):
    """Define a device."""

    id: int
    token: str
    platform: str
    endpoint: str
    created_at: datetime
    updated_at: datetime

    validate_created_at = validator("created_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )
    validate_updated_at = validator("updated_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )


class DeviceAllResponse(BaseModel):
    """Define an API response containing all devices."""

    devices: list[Device]


class DeviceGetResponse(BaseModel):
    """Define an API response containing a single device."""

    device: Device

    class Config:
        """Define model configuration."""

        fields = {
            "device": "devices",
        }



"""Define endpoints for interacting with bridges."""
from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any, cast

from aionotion.bridge.models import Bridge as BridgeModel
from aionotion.bridge.models import BridgeAllResponse, BridgeGetResponse
from aionotion.helpers.typing import BaseModelT


class Bridge:
    """Define an object to interact with all endpoints."""

    def __init__(
        self,
        request: Callable[..., Awaitable[dict[str, Any]]],
        request_and_validate: Callable[..., Awaitable[BaseModelT]],
    ) -> None:
        """Initialize.

        Args:
            request: The _request method from the Client.
            request_and_validated: The _request_and_validate method from the Client.
        """
        self._request = request
        self._request_and_validate = request_and_validate

    async def async_all(self) -> list[BridgeModel]:
        """Get all bridges.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            BridgeAllResponse,
            await self._request_and_validate("get", "base_stations", BridgeAllResponse),
        )
        return resp.bridges

    async def async_create(self, attributes: dict[str, Any]) -> BridgeModel:
        """Create a bridge with a specific attribute payload.

        Args:
            attributes: The attributes to assign to the new bridge.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            BridgeGetResponse,
            await self._request_and_validate(
                "post",
                "base_stations",
                BridgeGetResponse,
                json={"base_stations": attributes},
            ),
        )
        return resp.bridge

    async def async_delete(self, bridge_id: int) -> None:
        """Delete a bridge by ID.

        Args:
            bridge_id: The ID of the bridge to delete.
        """
        await self._request("delete", f"base_stations/{bridge_id}")

    async def async_get(self, bridge_id: int) -> BridgeModel:
        """Get a bridge by ID.

        Args:
            bridge_id: The ID of the bridge to get.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            BridgeGetResponse,
            await self._request_and_validate(
                "get", f"base_stations/{bridge_id}", BridgeGetResponse
            ),
        )
        return resp.bridge

    async def async_reset(self, bridge_id: int) -> BridgeModel:
        """Reset a bridge (clear its wifi credentials) by ID.

        Args:
            bridge_id: The ID of the bridge to reset.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            BridgeGetResponse,
            await self._request_and_validate(
                "put", f"base_stations/{bridge_id}/reset", BridgeGetResponse
            ),
        )
        return resp.bridge

    async def async_update(
        self, bridge_id: int, new_attributes: dict[str, Any]
    ) -> BridgeModel:
        """Update a bridge with a specific attribute payload.

        Args:
            bridge_id: The ID of the bridge to update.
            new_attributes: The new attributes to give the bridge.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            BridgeGetResponse,
            await self._request_and_validate(
                "put",
                f"base_stations/{bridge_id}",
                BridgeGetResponse,
                json={"base_stations": new_attributes},
            ),
        )
        return resp.bridge



"""Define bridge models."""
# pylint: disable=consider-alternative-union-syntax,too-few-public-methods
from __future__ import annotations

from datetime import datetime
from typing import Optional, Union

from pydantic import BaseModel, validator

from aionotion.helpers.validators import validate_timestamp


class FirmwareVersion(BaseModel):
    """Define firmware version info."""

    silabs: str
    wifi: str
    wifi_app: str


class Bridge(BaseModel):
    """Define a bridge."""

    id: int
    name: str
    mode: str
    hardware_id: str
    hardware_revision: int
    firmware_version: FirmwareVersion
    missing_at: Optional[datetime]
    created_at: datetime
    updated_at: datetime
    system_id: int
    firmware: FirmwareVersion
    links: dict[str, Union[int, str]]

    validate_missing_at = validator("missing_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )
    validate_created_at = validator("created_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )
    validate_updated_at = validator("updated_at", allow_reuse=True, pre=True)(
        validate_timestamp
    )


class BridgeAllResponse(BaseModel):
    """Define an API response containing all bridges."""

    bridges: list[Bridge]

    class Config:
        """Define model configuration."""

        fields = {
            "bridges": "base_stations",
        }


class BridgeGetResponse(BaseModel):
    """Define an API response containing a single bridge."""

    bridge: Bridge

    class Config:
        """Define model configuration."""

        fields = {
            "bridge": "base_stations",
        }



"""Define endpoints for interacting with devices (phones, etc. with Notion)."""
from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any, cast

from aionotion.device.models import Device as DeviceModel
from aionotion.device.models import DeviceAllResponse, DeviceGetResponse
from aionotion.helpers.typing import BaseModelT


class Device:
    """Define an object to interact with all endpoints."""

    def __init__(
        self,
        request: Callable[..., Awaitable[dict[str, Any]]],
        request_and_validate: Callable[..., Awaitable[BaseModelT]],
    ) -> None:
        """Initialize.

        Args:
            request: The _request method from the Client.
            request_and_validated: The _request_and_validate method from the Client.
        """
        self._request = request
        self._request_and_validate = request_and_validate

    async def async_all(self) -> list[DeviceModel]:
        """Get all devices.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            DeviceAllResponse,
            await self._request_and_validate("get", "devices", DeviceAllResponse),
        )
        return resp.devices

    async def async_create(self, attributes: dict[str, Any]) -> DeviceModel:
        """Create a device with a specific attribute payload.

        Args:
            attributes: The attributes to assign to the new device.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            DeviceGetResponse,
            await self._request_and_validate(
                "post", "devices", DeviceGetResponse, json={"devices": attributes}
            ),
        )
        return resp.device

    async def async_delete(self, device_id: int) -> None:
        """Delete a device by ID.

        Args:
            device_id: The ID of the device to delete.
        """
        await self._request("delete", f"devices/{device_id}")

    async def async_get(self, device_id: int) -> DeviceModel:
        """Get a device by ID.

        Args:
            device_id: The ID of the device to get.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            DeviceGetResponse,
            await self._request_and_validate(
                "get", f"devices/{device_id}", DeviceGetResponse
            ),
        )
        return resp.device



"""Define endpoints for interacting with systems (accounts)."""
from collections.abc import Awaitable, Callable
from typing import Any, cast

from aionotion.helpers.typing import BaseModelT
from aionotion.system.models import System as SystemModel
from aionotion.system.models import SystemAllResponse, SystemGetResponse


class System:
    """Define an object to interact with all endpoints."""

    def __init__(
        self,
        request: Callable[..., Awaitable[dict[str, Any]]],
        request_and_validate: Callable[..., Awaitable[BaseModelT]],
    ) -> None:
        """Initialize.

        Args:
            request: The request method from the Client object.
        """
        self._request = request
        self._request_and_validate = request_and_validate

    async def async_all(self) -> list[SystemModel]:
        """Get all systems.

        Returns:
            An API response payload.
        """
        resp = cast(
            SystemAllResponse,
            await self._request_and_validate("get", "systems", SystemAllResponse),
        )
        return resp.systems

    async def async_create(self, attributes: dict[str, Any]) -> SystemModel:
        """Create a system with a specific attribute payload.

        Args:
            attributes: The attributes to assign to the new system.

        Returns:
            An API response payload.
        """
        resp = cast(
            SystemGetResponse,
            await self._request_and_validate(
                "post", "systems", SystemGetResponse, json={"systems": attributes}
            ),
        )
        return resp.system

    async def async_delete(self, system_id: int) -> None:
        """Delete a system by ID.

        Args:
            system_id: The ID of the system to delete.
        """
        await self._request("delete", f"systems/{system_id}")

    async def async_get(self, system_id: int) -> SystemModel:
        """Get a system by ID.

        Args:
            system_id: The ID of the system to get.

        Returns:
            An API response payload.
        """
        resp = cast(
            SystemGetResponse,
            await self._request_and_validate(
                "get", f"systems/{system_id}", SystemGetResponse
            ),
        )
        return resp.system

    async def async_update(
        self, system_id: int, new_attributes: dict[str, Any]
    ) -> SystemModel:
        """Update a system with a specific attribute payload.

        Args:
            system_id: The ID of the system to update.
            new_attributes: The new attributes to give the system.

        Returns:
            An API response payload.
        """
        resp = cast(
            SystemGetResponse,
            await self._request_and_validate(
                "put",
                f"systems/{system_id}",
                SystemGetResponse,
                json={"systems": new_attributes},
            ),
        )
        return resp.system



"""Define endpoints for interacting with sensors."""
from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any, cast

from aionotion.helpers.typing import BaseModelT
from aionotion.sensor.models import Listener, ListenerAllResponse
from aionotion.sensor.models import Sensor as SensorModel
from aionotion.sensor.models import SensorAllResponse, SensorGetResponse


class Sensor:
    """Define an object to interact with all endpoints."""

    def __init__(
        self,
        request: Callable[..., Awaitable[dict[str, Any]]],
        request_and_validate: Callable[..., Awaitable[BaseModelT]],
    ) -> None:
        """Initialize.

        Args:
            request: The _request method from the Client.
            request_and_validated: The _request_and_validate method from the Client.
        """
        self._request = request
        self._request_and_validate = request_and_validate

    async def async_all(self) -> list[SensorModel]:
        """Get all sensors.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            SensorAllResponse,
            await self._request_and_validate("get", "sensors", SensorAllResponse),
        )
        return resp.sensors

    async def async_create(self, attributes: dict[str, Any]) -> SensorModel:
        """Create a sensor with a specific attribute payload.

        Args:
            attributes: The attributes to assign to the new sensor.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            SensorGetResponse,
            await self._request_and_validate(
                "post", "sensors", SensorGetResponse, json={"sensors": attributes}
            ),
        )
        return resp.sensor

    async def async_delete(self, sensor_id: int) -> None:
        """Delete a sensor by ID.

        Args:
            sensor_id: The ID of the sensor to delete.
        """
        await self._request("delete", f"sensors/{sensor_id}")

    async def async_get(self, sensor_id: int) -> SensorModel:
        """Get a sensor by ID.

        Args:
            sensor_id: The ID of the sensor to get.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            SensorGetResponse,
            await self._request_and_validate(
                "get", f"sensors/{sensor_id}", SensorGetResponse
            ),
        )
        return resp.sensor

    async def async_listeners(self) -> list[Listener]:
        """Get all listeners for all sensors.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            ListenerAllResponse,
            await self._request_and_validate(
                "get", "sensor/listeners", ListenerAllResponse
            ),
        )
        return resp.listeners

    async def async_listeners_for_sensor(self, sensor_uuid: str) -> list[Listener]:
        """Get all listeners for a sensor by UUID.

        Note that unlike other sensor endpoints, the sensor ID won't work here; the
        sensor UUID is required.

        Args:
            sensor_uuid: The UUID of the sensor to get.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            ListenerAllResponse,
            await self._request_and_validate(
                "get", f"sensors/{sensor_uuid}/listeners", ListenerAllResponse
            ),
        )
        return resp.listeners

    async def async_update(
        self, sensor_id: int, new_attributes: dict[str, Any]
    ) -> SensorModel:
        """Update a sensor with a specific attribute payload.

        Args:
            sensor_id: The ID of the sensor to update.
            new_attributes: The new attributes to give the sensor.

        Returns:
            A validated API response payload.
        """
        resp = cast(
            SensorGetResponse,
            await self._request_and_validate(
                "put",
                f"sensors/{sensor_id}",
                SensorGetResponse,
                json={"sensors": new_attributes},
            ),
        )
        return resp.sensor
